/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file ivector.hpp
///
/// @author The CAPD Group
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#include <stdexcept>

//---------------  definitions for class intervalVector only -------------------//

template<typename IVectorType>
IVectorType intervalHull(const IVectorType& A, const IVectorType& B)
{
   if(A.dimension()!=B.dimension())
      throw std::runtime_error("Incompatible vectors in function intervalHull");
   IVectorType v(A.dimension());
   typename IVectorType::iterator b=v.begin(), e=v.end();
   typename IVectorType::const_iterator b1=A.begin(), b2=B.begin();
   while(b!=e)
   {
      *b = intervalHull(*b1,*b2);
      ++b;
      ++b1;
      ++b2;
   }
   return v;
}


template<typename IVectorType>
void split(IVectorType& v, IVectorType& rv)
{
   if(v.dimension()!=rv.dimension())
      throw std::runtime_error("Incompatible vectors in function split");
   typename IVectorType::iterator b = v.begin(),e=v.end(), i=rv.begin();
   while(b!=e)
   {
      b->split(*i);
      ++b;
      ++i;
   }
}

template<typename IVectorType>
IVectorType midVector(const IVectorType& v)
{
   IVectorType temp(v.dimension());
   typename IVectorType::const_iterator b=v.begin(), e=v.end();
   typename IVectorType::iterator i = temp.begin();
   while(b!=e)
   {
      *i = b->mid();
      ++i;
      ++b;
   }
   return temp;
}

template<typename IVectorType>
IVectorType leftVector(const IVectorType& v)
{
   IVectorType temp(v.dimension());
   typename IVectorType::const_iterator b=v.begin(), e=v.end();
   typename IVectorType::iterator i = temp.begin();
   while(b!=e)
   {
      *i = b->leftBound();
      ++i;
      ++b;
   }
   return temp;
}

template<typename IVectorType>
IVectorType rightVector(const IVectorType& v)
{
   IVectorType temp(v.dimension());
   typename IVectorType::const_iterator b=v.begin(), e=v.end();
   typename IVectorType::iterator i = temp.begin();
   while(b!=e)
   {
      *i = b->rightBound();
      ++i;
      ++b;
   }
   return temp;
}

template<typename IVectorType>
typename IVectorType::ScalarType size(const IVectorType& v)
{
   typedef typename IVectorType::ScalarType ScalarType;
   ScalarType result(0.);
   typename IVectorType::const_iterator b=v.begin(), e=v.end();
   while(b!=e)
   {
      result = capd::max(result,ScalarType(b->rightBound()) - ScalarType(b->leftBound()));
      ++b;
   }
   return ScalarType(result.rightBound());
}

template<typename IVectorType>
IVectorType diam(const IVectorType& v)
{
   typedef typename IVectorType::ScalarType ScalarType;
   IVectorType result(v.dimension());
   typename IVectorType::iterator i=result.begin();
   typename IVectorType::const_iterator b=v.begin(), e=v.end();
   while(b!=e)
   {
      *i = (ScalarType(b->rightBound()) - ScalarType(b->leftBound())).rightBound();
      ++i;
      ++b;
   }
   return result;
}

template<typename IVectorType>
IVectorType intervalBall(const IVectorType &iv, const typename IVectorType::ScalarType &r)
{
   IVectorType result(iv.dimension());
   typename IVectorType::iterator b=result.begin(), e=result.end();
   typename IVectorType::const_iterator i=iv.begin();

   while(b!=e)
   {
      *b = ball(*i,r);
      ++b;
      ++i;
   }
   return result;
}

template<typename IVectorType>
typename IVectorType::ScalarType solveAffineInclusion(
      const IVectorType& a,
      const IVectorType& p,
      const IVectorType& c
   )
{
   typedef typename IVectorType::ScalarType ScalarType;

   if(a.dimension()!=p.dimension() || a.dimension()!=c.dimension())
      throw std::runtime_error("Incompatible vectors in function solveAffineInclusion");

   ScalarType result = solveAffineInclusion(a[0],p[0],c[0]);
   for(int i=1; i<a.dimension(); ++i)
   {
      ScalarType iv=solveAffineInclusion(a[i],p[i],c[i]);
      result = capd::min( result.leftBound(), iv.leftBound() );
   }
   return result;
}

template<typename IVectorType>
typename IVectorType::ScalarType solveAffineInclusion(
      const IVectorType& a,
      const IVectorType& p,
      const IVectorType& c,
      int& dir
   )
{
   if(a.dimension()!=p.dimension() || a.dimension()!=c.dimension())
      throw std::runtime_error("Incompatible vectors in function solveAffineInclusion");

   typedef typename IVectorType::ScalarType ScalarType;
   ScalarType result = solveAffineInclusion(a[0],p[0],c[0]);
   dir=0;

   for(int i=1; i<a.dimension(); ++i)
   {
      ScalarType iv = solveAffineInclusion(a[i],p[i],c[i]);
      if( iv.leftBound() < result.leftBound() )
      {
         result = iv.leftBound();
         dir = i;
      }
   }
   return result;
}


/******************************************************************************/
// checks if vector v1 is contained in vector v2
/******************************************************************************/

template<typename IVectorType>
bool subset(const IVectorType& v1, const IVectorType& v2)
{
   if(v1.dimension()!=v2.dimension())
      throw std::runtime_error("Incompatible vectors in function subset");
   typename IVectorType::const_iterator b = v1.begin(), e=v1.end(), i=v2.begin();
   while(b!=e)
   {
      if(!(b->subset(*i))) return false;
      ++b;
      ++i;
   }
   return true;
}

/******************************************************************************/
// checks if vector v1 is contained in interior of vector v2
/******************************************************************************/

template<typename IVectorType>
bool subsetInterior(const IVectorType& v1, const IVectorType& v2)
{
   if(v1.dimension()!=v2.dimension())
      throw std::runtime_error("Incompatible vectors in function subset");
   typename IVectorType::const_iterator b = v1.begin(), e=v1.end(), i=v2.begin();
   while(b!=e)
   {
      if(!(b->subsetInterior(*i))) return false;
      ++b;
      ++i;
   }
   return true;
}

/******************************************************************************/
// intersection of two intervalVectors
// halts if the intersection is empty
/******************************************************************************/

template<typename IVectorType>
IVectorType intersection(const IVectorType &v1, const IVectorType &v2)
{
   if(v1.dimension()!=v2.dimension())
      throw std::runtime_error("Incompatible vectors in function intersection");

   IVectorType result(v1.dimension());
   typename IVectorType::const_iterator b1 = v1.begin(), b2=v2.begin();
   typename IVectorType::iterator b = result.begin(), e=result.end();

   while(b!=e)
   {
      if( !intersection(*b1,*b2,*b))
      {
         throw std::runtime_error("Empty intersection of intervalVectors");
      }
      ++b;
      ++b1;
      ++b2;
   }
   return result;
}

/// @}
